"""
Provides a python interface to basic amarok DCOP functions.

Provies a layer that selects the "best" possible method of using dcop
with python based on:

- pydcop (Package: dcoppython on Gentoo, python-dcop on Debian )
- dcop cmdline client, as a subprocess (final fallback, included w/ amklib)


These modules don't always produce the same output.
The ezSetup(...) function works around these differences when doing a SELECT
query on the amarok collection.

Create your own bound functions with the newMethod(...) function.

"""

SETUP = None

DEBUG = 0

if not SETUP:
    try:
        import pydcop
        SETUP = 'pydcop'
    except:
        pass

if not SETUP:
    try:
        import pipedcop
        SETUP = 'pipedcop'
    except:
        pass


if not SETUP:
    raise ImportError, "No DCOP libraries available"



if 'pydcop' == SETUP:
   query = pydcop.DCOPMethod( 'amarok','collection', 'query' )
   append = pydcop.DCOPMethod( 'amarok', 'playlist', 'addMedia' )
   addMenuItem = pydcop.DCOPMethod( 'amarok', 'script', 'addCustomMenuItem' )
   delMenuItem = pydcop.DCOPMethod( 'amarok', 'script', 'removeCustomMenuItem' )
   
   _newMethod = pydcop.DCOPMethod


if 'pipedcop' == SETUP:
    query = pipedcop.mkcommand( 'amarok', 'collection', 'query' )
    append = pipedcop.mkcommand( 'amarok', 'playlist', 'addMedia' )

    addMenuItem = pipedcop.mkcommand( 'amarok', 'script', 'addCustomMenuItem' )
    delMenuItem = pipedcop.mkcommand('amarok','script','removeCustomMenuItem')

    _newMethod = pipedcop.mkcommand



def newMethod( app, module, function ):
    """Bind a DCOP method to a python function.
    Allows you to create DCOP function that aren't part of the limited list
    of functions defined in the mydcop module.

    This function should be dcop backend agnostic. However, running
    dcop via subprocesses versus a python dcop module may introduce 
    differences in the output of the functions. The query(...) function is
    a good example of this. It may be helpful to write a wrapper function
    that forces the behavior of the all possible modes to be the same.
    Take a look at the source of ezSelect to see one way to do this.
    """
    return _newMethod( app, module, function )



def ezSelect( columns, tables, where=None, limit=None, order=None, debug=0 ):
    """Wrap the basic queries into a format that's reasonably
    useful and hides some of the uglyness that querying the
    database over dcop seems to bring.
    
    Returns:
        an iterator containing a list of one item for each column

    Arguments:
        columns: a list of column names
        tables: a list or string containing table names
        where: a string holding a where clause (w/o the WHERE term)
        limit: a string holding the max number of rows to fetch
        order: a string holding a list of columns to sort by
    """

    templ = 'SELECT %s FROM %s %s %s %s'
    
    # set up addn'l query clauses
    if where:
        where = ' WHERE %s ' % where
    else:
        where = ''

    if limit:
        limit = ' LIMIT %s ' % limit
    else:
        limit = ''

    if order:
        order = ' ORDER BY %s ' % order
    else:
        order = ''

    
    # get the columns from a list or tuple
    numColumns = len( columns )
    columns = ', '.join( [ 'quote(%s)' % ss for ss in columns ] )


    #tables can be astring or list
    if type(tables) in ( list, tuple ):
        tables = ', '.join( tables )


    #create the final query string
    qs = templ % ( columns, tables, where, order, limit )

    if DEBUG or debug: print qs

    #execute the query, creating 'rows' using the expected number
    # of columns
    results = iter( query(qs) )

    try:
        while True:
            out = []
            for i in range( 0, numColumns ):
                out.append( deQuote( results.next() ) )
                

            yield out
        
    except StopIteration:
        pass
        
    return
    


def deQuote( txt ):
    if txt == '':
        return txt

    if  "'" in ( txt[0], txt[-1] ):
        return txt[1:-1].replace( "''", "'" )

    return txt
    

def dbQuote( value ):
    if type( value ) == str:
        return "'%s'" % ( value.replace( "'", "''" ) )

    return str(value)
