
"""
ScriptLoop deals with events
----------------------------

Use subclasses of the included ScriptLoop class to deal with 
the events amaroK writes to your scripts standard input. 
Create setUp, and tearDown functions for init and cleanup.
Create handle_NNN functions to deal with an amarok event with
the name NNN.
Create handleMenu_NNN function to deal with menu entries you've
created.


from amklib.scriptio import ScriptLoop

class MyLoop (ScriptLoop):
    def setUp(self):
        self.addMenuItem( "My Menu Item" )
        ...do other intialization...

    def handle_someEvent(self, argumentTxt):
        ...do whatever...

    def handleMenu_My_Menu_Item(self,url):
        ...do whatever for a menu click...

    def tearDown(self):
        self.clearMenuItem( "My Menu Item" )
        ...clean up other stuff...


myloop = MyLoop( "My Script Name" )
myloop.start() #waits for events from amarok




Standard Event Names
--------------------

- engineStateChange
- trackChange
- customMenuClicked: handled by built-in handle_customMenuClicked(...)

This is a cheat sheet. Not a comprehensive list :-P
"""

import re
import mydcop
import sys
import traceback
#import urllib

cmdregex = re.compile( '(?P<func>[^:]*)(?:: (?P<arg>([^:]* )*)(?P<url>.*))?' )


Commands = [ 'customMenuClicked' ]


SCRIPTNAME = 'XXX'
DEBUG = 0


STOP = 0
CONTINUE = 1

class ScriptLoop ( object ):
    """This class abstracts the main loop a amarok script that 
    listens to amarok status messages on the standard input usually
    needs to implement.

    It currently supports the built in amarok status messages
    (via subclassing and adding handler functions),
    plus allows you to define and react to custom menu items.

    What you want to do is sublcass this class and create handle_NNN 
    functions where NNN is equal to a valid amarok status event.

    By default you can also create/remove menu items and create
    handleMenu_NNN(...) functions.

    When start()
    """

    name = None  # used mainly for custom menu items
    io   = None  # the io input, typically stdin

    def __init__( self, appname, fio=None ):
        """Get all the state information ready.
        appname should be equal to whatever your script is called.
        It will appear as the sub-menu title when you right click
        on tracks.
        """
        self.name = appname

        if fio:
            self.io = fio
        else:
            self.io = sys.stdin
    
        self.menuRe = '(%s) (.*?) ([^ ]*)$' % re.escape(self.name)
        self.menuRe = re.compile( self.menuRe  )
        

    def addMenuItem( self, name ):
        """Wrapper to add a menu item
        """
        mydcop.addMenuItem( self.name, name ) 

    def clearMenuItem( self, name ):
        """Wrapper to remove a menu item
        """
        mydcop.delMenuItem( self.name, name )


    def parse_command( self, txt ):
        """Parse an amarok event into a command and argument
        """
        if ':' not in txt:
            return txt.strip(), ''

        return [ ss.strip() for ss in txt.split(':',1) ]
        


    def nameify( self, pre, txt ):
        """Convert a prepend string and general string into a
        string that is a valid function name.
        All adjacent non-alpha characters are converted into a single
        underscore. e.g 'View Information' -> 'View_Information'
        or 'Item Checkout...' -> 'Item_Checkout_'
        """
        
        txt = re.sub( '\W+', '_', txt ) 

        return '%s_%s' % ( pre, txt )


    def catchInput( self, txt ):
        """Handles each line from the input stream.
        Launches the various handler functions.
        Always checks for handle_all(line) first. Then handle_NNN(argument). 
        And then  handle_any(cmd,argument).

        If any of these functions return (amklib.scriptio.)CONTINUE
        processing will continue. Otherwise the function will return. 
        """
        if hasattr( self, 'handle_all' ):
            ret = self.handle_all(txt) 
            if not ret: 
                return STOP

        cmd, txt = self.parse_command( txt )

        fname = 'handle_%s' % cmd.replace( ' ', '_' ).replace( '-', '_' )

        if hasattr( self, fname ):
            ret = getattr( self, fname )( txt )
        else:
            ret = CONTINUE

        if not ret:
            return STOP

        if hasattr( self, 'handle_any' ):
            return self.handle_any( cmd, txt )
            
        return
            

    def handle_customMenuClicked( self, txt ):
        """Handles custom Menu Item events.
        It calls appropriate handleMenu_NNN(...) functions where
        NNN is the simplified name of a menu item.

        It is pretty useful as-is, but nothing stops you from overriding
        it in a subclass.
        """
        if DEBUG: print txt
        match = self.menuRe.search(txt)
        
        if not match:
            return CONTINUE

        name, fname, url = match.groups()

        fname = self.nameify( 'handleMenu',fname )

        if hasattr( self, fname ):
            return getattr( self, fname )( url )
        else:
            return CONTINUE
        
        
        

    def handle_any( self, cmd, txt ):
        """General catch-all function for unhandled
        events
        """
        if DEBUG: print 'Fall-through:', cmd, '-', txt
        return STOP


    def start( self ):
        """Begin processing the event loop. Doesn't do much if
        you haven't subclassed. :-)

        Create handle_NNN( arg ) member functions to handle events.
        Create handleMenu_NNN( url ) member functions to handle menu events.
        
        Create setUp() member function to initialize.
        Crete tearDown() member function to handle the end of processing.
        """
        if hasattr( self, 'setUp' ):
            self.setUp()

        try:
        
            while True:

                line = self.io.readline()
                
                if not line:
                    break
                else:
                    self.catchInput( line )


        except Exception, excp:
            if hasattr( self, 'tearDown' ):
                self.tearDown()

            traceback.print_exc()
            sys.exit( 1 )
            #raise excp.__class__, excp
            

        if hasattr( self, 'tearDown' ):
            self.tearDown()




__oldjunk__="""


def parseLine( strdata ):
    strdata = strdata.strip() + ' '  #strange, but it makes it work
    m = cmdregex.search( strdata )

    if not m:
        return ( None, None, None )


    m = m.groupdict()

    func = m['func']
    if type(func) == str: func = func.strip()

    arg = m['arg']
    if type(arg) == str: arg = arg.strip()

    url = m['url']
    if type(url) == str: url = url.strip()

    return ( func, arg, url )



def guardCmd( command ):
    func,arg,url = command
    if func not in Commands:
        return True

    if func == 'customMenuClicked' and SCRIPTNAME not in arg:
        return True

    return False



def addCommand( name ):
    mydcop.addMenuItem( SCRIPTNAME, name )

def delCommand( name ):
    mydcop.delMenuItem( SCRIPTNAME, name )



def checkCustomCommand( setup, arg ):
    for key,item in setup.iteritems():
        if DEBUG: print arg
        ok = True
        for word in item:
            if word not in arg.lower():
                ok = False

        if ok: return key

    return None

"""




